// bmp2asm.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#define _RGB2Lum(r,g,b) \
	((((long)(r) * 299) + ((long)(g) * 587) + ((long)(b) * 114)) / 1000)
#define IS_WIN30_DIB(lpbi)  ((*(LPDWORD)(lpbi)) == sizeof(BITMAPINFOHEADER))
#define WIDTHBYTES(bits)    (((bits) + 31) / 32 * 4)
#define DIB_HEADER_MARKER   ((WORD) ('M' << 8) | 'B')

// global config
BOOL g_bHex = FALSE;
BOOL g_bMask = FALSE;
BOOL g_bIcon = FALSE;
LPSTR g_szFile = NULL;
////

int usage()
{
	puts(
		"BMP2ASM v2.0, by <neuros666@yahoo.com>\n"
		"\n"
		"Usage:\tBMP2ASM [-i] [-m] <bmpfile>\n"
		"\n"
		"BMP2ASM opens the specified bitmap file and prints \n"
		"the resulting assembly code to the standard output. \n"
		"The code generated is compatible with the aslc86k \n"
		"assembler v1.9 written by <marcus@idonex.se>. \n"
		"\n"
		"options \n"
		"------- \n"
		"-i\tgenerate icon data (default is sprite) \n"
		"-m\tgenerate sprite mask (sprite mode only) \n"
		"\n"
		"icon mode \n"
		"--------- \n"
		"The icon mode supports indexed color bitmaps only \n"
		"(1, 2, 4 or 8 bpp). It generates the icon palette \n"
		"with the first 16 colors of the palette, or fills \n"
		"the remainder with solid white if the bitmap contains \n"
		"less than 16 colors. The icon data is generated from \n"
		"the 32x32 pixels at the top-left corner, regardless \n"
		"of the actual size of the bitmap. An error occurs if \n"
		"the bitmap contains no palette or has less than 32 \n"
		"pixels either in width or height. The mask option is \n"
		"ignored in icon mode. \n"
		"\n"
		"sprite mode \n"
		"----------- \n"
		"The sprite mode supports bitmaps of 1, 2, 4, 8 or \n"
		"24 bpp. It generates the sprite data according to \n"
		"the luminance of each pixel (>128 == 1; <128 == 0). \n"
		"The resulting sprite has the same dimensions as the \n"
		"source bitmap. The mask option assumes that the \n"
		"bottom half of the bitmap contains mask data that \n"
		"tell which pixels in the sprite above are to be \n"
		"solid and which are to be transparent (writing a \n"
		"sprite routine that uses such a mask is left to the \n"
		"user). An error occurs if the bitmap height is not \n"
		"multiple of 2. \n"
		"\n"
		"conformance \n"
		"----------- \n"
		"All code generated is conform to the tetris source \n"
		"code as written by <marcus@idonex.se>. \n"
		);
	return 0;
}

HANDLE read_dib(HANDLE hFile)
{
	DWORD dwRead;

	BITMAPFILEHEADER bmfhdr;
	if (!::ReadFile(hFile, &bmfhdr, sizeof(BITMAPFILEHEADER), &dwRead, NULL))
	{
		puts("\t;; ERROR : read_dib() : can't read bitmap file header!\n");
		return NULL;
	}

	if (bmfhdr.bfType != DIB_HEADER_MARKER)
	{
		puts("\t;; ERROR : read_dib() : invalid bitmap file signature!\n");
		return NULL;
	}

	DWORD dwSize = ::GetFileSize(hFile, NULL) - sizeof(BITMAPFILEHEADER);
	HANDLE hDIB = ::GlobalAlloc(GHND, dwSize);
	if (!hDIB)
	{
		puts("\t;; ERROR : read_dib() : can't allocate bitmap memory!\n");
		return NULL;
	}

	LPSTR lpDIB = (LPSTR)::GlobalLock(hDIB);
	if (!lpDIB)
	{
		puts("\t;; ERROR : read_dib() : can't lock bitmap memory!\n");
		::GlobalFree(hDIB);
		return NULL;
	}

	if (!::ReadFile(hFile, lpDIB, dwSize, &dwRead, NULL))
	{
		puts("\t;; ERROR : read_dib() : can't read bitmap file data!\n");
		::GlobalUnlock(hDIB);
		::GlobalFree(hDIB);
		return NULL;
	}

	if (!IS_WIN30_DIB(lpDIB))
	{
		puts("\t;; ERROR : read_dib() : can't process bitmaps anterior to version 3.0!\n");
		::GlobalUnlock(hDIB);
		::GlobalFree(hDIB);
		return NULL;
	}

	::GlobalUnlock(hDIB);
	return hDIB;
}

size_t colors_used(LPBITMAPINFOHEADER lpBI)
{
	if (lpBI->biBitCount > 8)
		return 0;
	if (lpBI->biClrUsed)
		return lpBI->biClrUsed;
	return (1 << lpBI->biBitCount);
}

#define DUMP_DEFAULT	(0x0)

#define DUMP_BIN		(0x0)
#define DUMP_HEX		(0x1)

#define DUMP_BYTE		(0x0)
#define DUMP_WORD		(0x2)

void dump_buffer(LPBYTE buf, size_t len, int flags)
{
	if (flags & DUMP_WORD)
	{
		len >>= 1;
		LPWORD lpwbuf = (LPWORD)buf;
		printf("\t.word\t");
		for (; len--; lpwbuf++)
		{
			printf("$%04X", *lpwbuf);
			if (len) putchar(',');
		}
	}
	else	// byte
	{
		printf("\t.byte\t");
		if (flags & DUMP_HEX)
		{
			for (; len--; buf++)
			{
				printf("$%02X", *buf);
				if (len) putchar(',');
			}
		}
		else
		{
			BYTE unit = 0;
			for (; len--; buf++)
			{
				unit = (BYTE)*buf;
				putchar('%');
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0'); unit <<= 1;
				putchar(unit & 0x80 ? '1' : '0');
				if (len) putchar(',');
			}
		}
	}
	putchar('\n');
}

// global DIB data
LPBITMAPINFOHEADER g_lpBI = NULL;
LPRGBQUAD g_lpPal = NULL;
LPSTR g_lpSurface = NULL;
////

size_t spr_crunch_scanline(LPSTR lpLine, LPBYTE buf, size_t maxlen)
{
	LPSTR curr_pel = lpLine;
	LPBYTE curr_byte = buf;
	int curr_bit = 0;
	memset(buf, 0, maxlen);

	switch (g_lpBI->biBitCount)
	{
	// RGB
	case 24:
		{
			curr_bit = 8;

			for (LONG i = g_lpBI->biWidth; i--; )
			{
				curr_bit--;
				if (!(_RGB2Lum((unsigned)curr_pel[0], (unsigned)curr_pel[1], (unsigned)curr_pel[2]) & 0x80))
				{
					*curr_byte |= (1 << curr_bit);
				}
				if (!curr_bit) 
				{
					++curr_byte;
					curr_bit = 8;
				}
				curr_pel += 3;
			}
		}
		break;

	// indexed
	case 8:
	case 4:
	case 2:
	case 1:
		{
			int index = 0;
			int ppb = 8 / g_lpBI->biBitCount;	// pixels-per-byte

			curr_bit = 8;
			int pel_in_byte = ppb;

			BYTE mask = (1 << g_lpBI->biBitCount) - 1;
			for (LONG i = g_lpBI->biWidth; i--; )
			{
				curr_bit--;
				pel_in_byte--;

				index = (unsigned)(*curr_pel >> (pel_in_byte * g_lpBI->biBitCount)) & mask;
				if (!(_RGB2Lum(g_lpPal[index].rgbRed, g_lpPal[index].rgbGreen, g_lpPal[index].rgbBlue) & 0x80))
				{
					*curr_byte |= (1 << curr_bit);
				}

				if (!curr_bit) 
				{
					++curr_byte;
					curr_bit = 8;
				}
				if (!pel_in_byte) 
				{
					++curr_pel;
					pel_in_byte = ppb;
				}
			}
		}
		break;

	default:
		puts("\t;; ERROR : crunch_scanline() : unsupported bit count\n");
		return 0;
	}

	return (size_t)(curr_byte - buf + (curr_bit % 8 ? 1 : 0));
}

int spr_generate()
{
	// validate mask request
	if (g_bMask && (g_lpBI->biHeight & 1))
	{
		puts("\t;; ERROR : main() : mask requested for a bitmap with an odd height!\n");
		return 0;
	}

	// create label name
	{
		LPSTR p = NULL;
		
		// remove file path
		p = strrchr(g_szFile, '\\');
		if (p) g_szFile = p + 1;
		
		// remove file extension
		p = strrchr(g_szFile, '.');
		if (p) *p = 0;
		
		// replace spaces for underscores + convert to lowercase
		for (int i = strlen(g_szFile); i--; )
		{
			if (isalpha(g_szFile[i]))
				g_szFile[i] |= 0x20;
			if (strchr("\x08\x09\x0A\x0C\x0D\x20", g_szFile[i]))
				g_szFile[i] = '_';
		}
	}

	printf(
		"%s:\n"
		"\t.byte\t%u,%u\n",
		g_szFile, (unsigned char)g_lpBI->biWidth, 
		(unsigned char)(g_bMask ? g_lpBI->biHeight >> 1 : g_lpBI->biHeight));

	// crunch the pixels!
	LPSTR lpLine = g_lpSurface;
	DWORD dwPitch = WIDTHBYTES(g_lpBI->biWidth * g_lpBI->biBitCount);
	size_t len, maxlen = (g_lpBI->biWidth + 7) >> 3;
	LPBYTE buf = new BYTE [maxlen];
	LONG h = g_lpBI->biHeight;
	for (lpLine += g_lpBI->biHeight * dwPitch; h--; )
	{
		lpLine -= dwPitch;
		len = spr_crunch_scanline(lpLine, buf, maxlen);
		if (!len) break;
		dump_buffer(buf, len, DUMP_BIN | DUMP_BYTE);
		
		if (g_bMask && h == (g_lpBI->biHeight >> 1))	// remaining half is the mask -- dump header
			printf("%s_mask:\n", g_szFile);
	}
	delete [] buf;
	return 1;
}

size_t ico_crunch_scanline(LPSTR lpLine, LPBYTE buf, size_t maxlen)
{
	LPSTR curr_pel = lpLine;
	LPBYTE curr_byte = buf;
	int curr_bit = 0;
	memset(buf, 0, maxlen);

	switch (g_lpBI->biBitCount)
	{
	case 24:	// error!
		return 0;

	case 8:
	case 4:
	case 2:
	case 1:
		{
			int index = 0;
			int ppb = 8 / g_lpBI->biBitCount;	// pixels-per-byte

			curr_bit = 8;
			int pel_in_byte = ppb;

			BYTE mask = (1 << g_lpBI->biBitCount) - 1;
			for (LONG i = 32; i--; )
			{
				curr_bit -= 4;
				pel_in_byte--;

				index = (unsigned)(*curr_pel >> (pel_in_byte * g_lpBI->biBitCount)) & mask;
				*curr_byte |= ((index & 0xF) << curr_bit);

				if (!curr_bit) 
				{
					++curr_byte;
					curr_bit = 8;
				}
				if (!pel_in_byte) 
				{
					++curr_pel;
					pel_in_byte = ppb;
				}
			}
		}
		break;

	default:
		puts("\t;; ERROR : crunch_scanline() : unsupported bit count\n");
		return 0;
	}

	return (size_t)(curr_byte - buf + (curr_bit % 8 ? 1 : 0));
}

int ico_generate()
{
	// make sure the bitmap is 8 bpp or less
	if (g_lpBI->biBitCount > 8)
	{
		puts("\t;; ERROR : ico_generate() : bitmap has no palette (not 8 bpp or less)!\n");
		return 0;
	}

	// make sure the bitmap is 8 bpp or less
	if (g_lpBI->biWidth < 32 || g_lpBI->biHeight < 32)
	{
		puts("\t;; ERROR : ico_generate() : bitmap has less than 32 in either dimension!\n");
		return 0;
	}

	printf(
		"\t;; Icon header\n"
		"\n"
		"\t.org\t$240\n"
		"\n"
		"\t.word\t1,10\n"
		"\n"
		"\t;; Icon palette\n"
		"\n"
		"\t.org\t$260\n"
		"\n"
		);

	size_t len, maxlen = 32;
	LPBYTE buf = new BYTE [maxlen];

	// crunch the palette
	{
		size_t colors = colors_used(g_lpBI);
		LPWORD lpwbuf = (LPWORD)buf + 16;
		for (size_t i = 16; i--; )
		{
			--lpwbuf;
			if (i < colors) // get from palette
			{
				*lpwbuf = 0xF000 |
					((g_lpPal[i].rgbRed   & 0xF0) << 4) |
					((g_lpPal[i].rgbGreen & 0xF0)) |
					((g_lpPal[i].rgbBlue  & 0xF0) >> 4);
			}
			else	// fill with solid white
			{
				*lpwbuf = 0xFFFF;
			}
		}
		dump_buffer(buf, 16, DUMP_HEX | DUMP_WORD);
		dump_buffer(buf + 16, 16, DUMP_HEX | DUMP_WORD);
	}


	printf(
		"\n"
		"\t;; Icon \n"
		"\n"
		"\t.org\t$280\n"
		"\n"
		);

	// crunch the pixels!
	LPSTR lpLine = g_lpSurface;
	DWORD dwPitch = WIDTHBYTES(g_lpBI->biWidth * g_lpBI->biBitCount);
	LONG h = 32;
	for (lpLine += g_lpBI->biHeight * dwPitch; h--; )
	{
		lpLine -= dwPitch;
		len = ico_crunch_scanline(lpLine, buf, maxlen);
		if (!len) break;
		dump_buffer(buf, len, DUMP_HEX | DUMP_BYTE);
	}
	delete [] buf;
	return 1;
}

int parse_cmdline(int argc, char* argv[])
{
	while (--argc)
	{
/*		if (!stricmp(argv[argc], "-X"))
		{
			g_bHex = TRUE;
		}
		else
*/		if (!stricmp(argv[argc], "-M"))
		{
			g_bMask = TRUE;
		}
		else
		if (!stricmp(argv[argc], "-I"))
		{
			g_bIcon = TRUE;
		}
		else
		{
			g_szFile = argv[argc];
		}
	}
	return 1;
}

int find_DIB_data(LPSTR lpDIB)
{
	g_lpBI = (LPBITMAPINFOHEADER)lpDIB;
	if (g_lpBI->biCompression)
	{
		puts("\t;; ERROR : main() : RLE compression not supported!\n");
		return 0;
	}
	g_lpPal = (LPRGBQUAD)((LPSTR)g_lpBI + sizeof(BITMAPINFOHEADER));
	g_lpSurface = NULL;
	{
		size_t n = colors_used(g_lpBI);
		if (n)
		{
			g_lpSurface = (LPSTR)(g_lpPal + n);
		}
		else
		{
			g_lpSurface = (LPSTR)g_lpPal;
			g_lpPal = NULL;
		}
	}
	return 1;
}

int main(int argc, char* argv[])
{
	if (argc == 1)
		return usage();
	if (!parse_cmdline(argc, argv))
		return 0;

	puts("\t;; Created with BMP2ASM v2.0, by <neuros666@yahoo.com>");

	HANDLE hFile = INVALID_HANDLE_VALUE;
	printf("\t;; Source file: \"%s\"\n\n", g_szFile);
	hFile = ::CreateFile(g_szFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		puts("\t;; ERROR : main() : can't open bitmap file!\n");
		return 0;
	}

	HANDLE hDIB = read_dib(hFile);
	::CloseHandle(hFile);

	if (!hDIB)
		return 0;

	LPSTR lpDIB = (LPSTR)::GlobalLock(hDIB);
	if (!lpDIB)
	{
		puts("\t;; ERROR : main() : can't lock bitmap memory!\n");
		::GlobalFree(hDIB);
		return 0;
	}

	if (!find_DIB_data(lpDIB))
	{
		::GlobalUnlock(hDIB);
		::GlobalFree(hDIB);
		return 0;
	}

	if (g_bIcon)
		ico_generate();
	else
		spr_generate();

	::GlobalUnlock(hDIB);
	::GlobalFree(hDIB);
//	fflush(stdin);getchar();
	return 0;
}
